@@ -2,8 +2,8 @@
 
 __authors__ = ( 'Fabio Cumbo (fabio.cumbo@unitn.it)',
                 'Simone Truglia (s.truglia@students.uninettunouniversity.net)' )
-__version__ = '1.0.5'
-__date__ = 'Apr 20, 2022'
+__version__ = '1.0.7'
+__date__ = 'Jan 27, 2023'
 
 import sys
 
@@ -21,7 +21,7 @@ import chopin2.functions as fun
 def cite():
     """
     Custom argparse action to print citations and exit
-    Usage: python chopin2.py --cite
+    Usage: chopin2 --cite
     """
     
     class citeAction(ap.Action):
@@ -36,19 +36,40 @@ def cite():
             parser.exit()
     return citeAction
 
+def number(typev, minv=None, maxv=None):
+    """
+    Take full control of input numeric types by defining custom intervals
+    """
+
+    def type_func(value):
+        """
+        Test data type and ranges on the input value
+        """
+
+        try:
+            value = typev(value)
+            if minv and value < minv:
+                raise ap.ArgumentTypeError("Minimum value is {}".format(minv))
+            if maxv and value > maxv:
+                raise ap.ArgumentTypeError("Maximum value is {}".format(maxv))
+            return value
+        except:
+            raise ap.ArgumentTypeError("Input value must be {}".format(typev))
+
+    return type_func
+
 def read_params():
-    p = ap.ArgumentParser( description = ( "The chopin2.py script builds and tests a Hyperdimensional Computing "
-                                           "classification model on a input dataset." ),
+    p = ap.ArgumentParser( description = ( "Supervised Classification with Hyperdimensional Computing." ),
                            formatter_class = ap.ArgumentDefaultsHelpFormatter )
     p.add_argument( '--dimensionality', 
-                    type = int,
+                    type = number(int, minv=10),
                     default = 10000,
                     help = "Dimensionality of the HD model" )
     p.add_argument( '--levels', 
-                    type = int,
+                    type = number(int, minv=2),
                     help = "Number of level hypervectors" )
     p.add_argument( '--retrain', 
-                    type = int,
+                    type = number(int, minv=0),
                     default = 0,
                     help = "Number of retraining iterations" )
     p.add_argument( '--stop',
@@ -62,29 +83,23 @@ def read_params():
                     type = str,
                     default = ',',
                     help = "Field separator" )
-    p.add_argument( '--training', 
-                    type = float,
-                    default = 80.0,
+    p.add_argument( '--psplit_training', 
+                    type = number(float, minv=0.0, maxv=100.0),
+                    default = 0.0,
                     help = ( "Percentage of observations that will be used to train the model. "
                              "The remaining percentage will be used to test the classification model" ) )
-    p.add_argument( '--k_folds', 
-                    type = int,
-                    default = 1,
-                    help = ( "Number of folds for cross validation. "
-                             "Cross validate HD models if --k_folds greater than 1" ) )
-    p.add_argument( '--auto_folds', 
-                    action = 'store_true',
-                    default = False,
-                    help = ( "Automatically estimate the number of folds for cross validation. "
-                             "It automatically override --k_folds" ) )
-    p.add_argument( '--seed', 
-                    type = int,
+    p.add_argument( '--crossv_k', 
+                    type = number(int, minv=2),
                     default = 0,
-                    help = ( "Seed for reproducing random sampling of the observations in the dataset "
-                             "and build both the training and test set" ) )
+                    help = ( "Number of folds for cross validation. "
+                             "Cross validate HD models if --crossv_k greater than 1" ) )
+    p.add_argument( '--seed', 
+                    type = number(int),
+                    default = 0,
+                    help = "Random seed" )
     p.add_argument( '--pickle', 
                     type = str,
-                    help = "Path to the pickle file. If specified, \"--dataset\", \"--fieldsep\", and \"--training\" parameters are not used" )
+                    help = "Path to the pickle file. If specified, \"--dataset\" and \"--fieldsep\" parameters are not used" )
     p.add_argument( '--features', 
                     type = str,
                     help = "Path to a file with a single column containing the whole set or a subset of feature" )
@@ -92,9 +107,9 @@ def read_params():
                     action = 'store_true',
                     default = False,
                     help = ( "This triggers the backward variable selection method for the identification of the most significant features. " 
-                             "Warning: computationally intense!") )
+                             "Warning: computationally intense!" ) )
     p.add_argument( '--group_min', 
-                    type = int,
+                    type = number(int, minv=1),
                     default = 1,
                     help = "Minimum number of features among those specified with the --features argument" )
     p.add_argument( '--dump', 
@@ -110,21 +125,21 @@ def read_params():
                     default = False,
                     help = "Do not delete the level hypervectors. It works in conjunction with --cleanup only" )
     p.add_argument( '--accuracy_threshold', 
-                    type = float,
+                    type = number(float, minv=0.0, maxv=100.0),
                     default = 60.0,
                     help = "Stop the execution if the best accuracy achieved during the previous group of runs is lower than this number" )
     p.add_argument( '--accuracy_uncertainty_perc', 
-                    type = float,
+                    type = number(float, minv=0.0, maxv=100.0),
                     default = 5.0,
-                    help = ( "Take a run into account even if its accuracy is lower than the best accuracy achieved in the same group minus "
-                             "its \"accuracy_uncertainty_perc\" percent" ) )
+                    help = ( "Take a run into account if its accuracy is lower than the best accuracy achieved in the same group, but greater"
+                             "than the best accuracy minus its \"accuracy_uncertainty_perc\" percent" ) )
     # Apache Spark
     p.add_argument( '--spark',
                     action = 'store_true',
                     default = False,
                     help = "Build the classification model in a Apache Spark distributed environment" )
     p.add_argument( '--slices', 
-                    type = int,
+                    type = number(int, minv=0),
                     help = ( "Number of slices in case --spark argument is enabled. "
                              "This argument is ignored if --gpu is enabled" ) )
     p.add_argument( '--master', 
@@ -141,13 +156,13 @@ def read_params():
                     help = ( "Build the classification model on an NVidia powered GPU. "
                              "This argument is ignored if --spark is specified" ) )
     p.add_argument( '--tblock', 
-                    type = int,
+                    type = number(int, minv=0),
                     default = 32,
                     help = ( "Number of threads per block in case --gpu argument is enabled. "
                              "This argument is ignored if --spark is enabled" ) )
     # Standard multiprocessing
     p.add_argument( '--nproc', 
-                    type = int,
+                    type = number(int, minv=1, maxv=os.cpu_count()),
                     default = 1,
                     help = ( "Number of parallel jobs for the creation of the HD model. "
                              "This argument is ignored if --spark is enabled" ) )
@@ -187,7 +202,7 @@ def chopin2():
         # If the pickle file already exists
         picklepath = args.pickle
         dataset = pickle.load(open(picklepath, 'rb'))
-        if len(dataset) > 4:
+        if len(dataset) > 2:
             # Define features, allData, and allLabels
             features, allData, allLabels = dataset
         else:
@@ -216,18 +231,24 @@ def chopin2():
         pickle.dump( pickledata, open( picklepath, 'wb' ) )
 
     if features and allData and allLabels:
-        if args.k_folds < 1 or args.k_folds > len(allData):
-            raise Exception("The number of folds under --k_folds must be greater than 1 and lower than the number of observations in the training set")
-        k_folds = args.k_folds
-        if args.auto_folds:
-            # Automatically estimate the number of folds for cross validation
-            N = float(len(allData))
-            test_perc = (100.0-args.training)/100.0
-            k_folds = math.ceil(N/(N*test_perc))
-        
-        if k_folds == 1:
+        if args.psplit_training == 0.0 and args.crossv_k == 0:
+            raise Exception("Please specify one of the following arguments: [\"--psplit_training\", \"--crossv_k\"]")
+
+        if args.psplit_training > 0.0:
             # Prepare training and test dataset with percentage split
-            trainData, trainLabels, testData, testLabels = fun.percentage_split(allData, allLabels, training=args.training, seed=args.seed)
+            trainData, trainLabels, testData, testLabels = fun.percentage_split(allData, allLabels, training=args.psplit_training, seed=args.seed)
+        elif args.crossv_k > 0:
+            # Split training set into k subsets
+            subsets = list()
+            dataSplit = [(allData[i::args.crossv_k], allLabels[i::args.crossv_k]) for i in range(args.crossv_k)]
+            for testData, testLabels in dataSplit:
+                trainData = list()
+                trainLabels = list()
+                for idx, subdata in enumerate(allData):
+                    if subdata not in testData:
+                        trainData.append(subdata)
+                        trainLabels.append(allLabels[idx])
+                subsets.append( (trainData, trainLabels, testData, testLabels) )
 
         # Take track of the best selected features 
         last_best_group = None
@@ -271,13 +292,13 @@ def chopin2():
             summary_filepath = os.path.join( datasetdir, 'summary.txt' )
             if not os.path.exists( summary_filepath ):
                 with open( summary_filepath, 'a+' ) as summary:
-                    header = '# Dataset: {}\n'.format( picklepath )
+                    header = '# Dataset: {}\n'.format( os.path.splitext(os.path.basename(args.dataset))[0] )
                     header += '# Dimensionality: {}\n'.format( args.dimensionality )
                     header += '# Number of levels: {}\n'.format( args.levels )
                     header += '# Max retraining iterations: {}\n'.format( args.retrain )
                     header += '# Accuracy threshold (stop condition): {}\n'.format( args.accuracy_threshold )
                     header += '# Accuracy uncertainty (percentage): {}\n'.format( args.accuracy_uncertainty_perc )
-                    header += '# Run ID\tGroup Size\tRetraining\tAccuracy'
+                    header += '# Run ID\tGroup Size\tRetraining\tAccuracy\tExcluded Feature'
                     fun.printlog( header, out=summary )
         # For each group size
         prev_group_size = np.Inf
@@ -309,6 +330,7 @@ def chopin2():
                     # Define a set of N features with N equals to "group_size"
                     for comb_features in itertools.combinations( best_features, group_size ):
                         comb_features = sorted(comb_features)
+                        excluded_features = list(set(best_features).difference(set(comb_features)))
                         # Build unique identifier for the current set of features
                         features_hash = hashlib.md5(str(comb_features).encode()).hexdigest()
                         # Create a log for the current run
@@ -342,33 +364,31 @@ def chopin2():
                         # Could be more than 1 in case of cross validation
                         models = list()
                         t0model = time.time()
-                        if k_folds > 1:
+                        
+                        if args.crossv_k > 0:
+                            # Run k-fold cross-validation
                             fun.printlog( 
-                                '\t--k_folds {}'.format(k_folds),
+                                '\t--crossv_k {}'.format(args.crossv_k),
                                 verbose=args.verbose,
                                 out=run
                             )
-                            # Reshape dataset if required
-                            allData_subset = allData
-                            if len(comb_features) < len(features):
-                                allData_subset = [[value for index, value in enumerate(obs) if features_idx[index]] for obs in allData]
-                            # Split training set into k subsets
-                            dataSplit = [(allData_subset[i::k_folds], allLabels[i::k_folds]) for i in range(k_folds)]
-                            for k, (testData, testLabels) in enumerate(dataSplit):
-                                trainData = list()
-                                trainLabels = list()
-                                for idx, subdata in enumerate(allData_subset):
-                                    if subdata not in testData:
-                                        trainData.append(subdata)
-                                        trainLabels.append(allLabels[idx])
+                            # Iterate over k subsets
+                            for k, (trainData, trainLabels, testData, testLabels) in enumerate(subsets):
+                                # Reshape dataset by removing features if needed
+                                trainData_subset = trainData
+                                testData_subset = testData
+                                if len(comb_features) < len(features):
+                                    trainData_subset = [[value for index, value in enumerate(obs) if features_idx[index]] for obs in trainData]
+                                    testData_subset = [[value for index, value in enumerate(obs) if features_idx[index]] for obs in testData]
+
                                 # Build the HD model
-                                model = fun.buildHDModel( trainData, 
+                                model = fun.buildHDModel( trainData_subset, 
                                                           trainLabels, 
-                                                          testData, 
+                                                          testData_subset, 
                                                           testLabels, 
                                                           args.dimensionality, 
                                                           args.levels, 
-                                                          os.path.splitext( os.path.basename( picklepath ) )[0],
+                                                          os.path.splitext(os.path.basename(picklepath))[0],
                                                           features_hash,
                                                           workdir=group_dir,
                                                           levelsdir=datasetdir,
@@ -380,14 +400,18 @@ def chopin2():
                                                           gpu=args.gpu,
                                                           tblock=args.tblock,
                                                           nproc=args.nproc,
-                                                          verbose=(args.verbose and k_folds == 1),
+                                                          verbose=False,
                                                           log=run,
                                                           seed=args.seed
                                                         )
                                 models.append(model)
-
                         else:
-                            # Use the precomputed training and test set with percentage split
+                            # Run k-fold cross-validation
+                            fun.printlog( 
+                                '\t--psplit_training {}'.format(args.psplit_training),
+                                verbose=args.verbose,
+                                out=run
+                            )
                             # Reshape dataset if required
                             trainData_subset = trainData
                             testData_subset = testData
@@ -401,7 +425,7 @@ def chopin2():
                                                       testLabels, 
                                                       args.dimensionality, 
                                                       args.levels, 
-                                                      os.path.splitext( os.path.basename( picklepath ) )[0],
+                                                      os.path.splitext(os.path.basename(picklepath))[0],
                                                       features_hash,
                                                       workdir=group_dir,
                                                       levelsdir=datasetdir,
@@ -412,7 +436,7 @@ def chopin2():
                                                       gpu=args.gpu,
                                                       tblock=args.tblock,
                                                       nproc=args.nproc,
-                                                      verbose=(args.verbose and k_folds == 1),
+                                                      verbose=args.verbose,
                                                       log=run,
                                                       seed=args.seed
                                                     )
@@ -420,7 +444,7 @@ def chopin2():
                         
                         t1model = time.time()
                         fun.printlog( 
-                            'Total elapsed time (model/s) {}s\nRetraining the HD model/s {} times at most'.format( int( t1model - t0model ), args.retrain ),
+                            'Total elapsed time (building) {}s'.format(int(t1model-t0model)),
                             verbose=args.verbose,
                             out=run
                         )
@@ -440,8 +464,8 @@ def chopin2():
                                                                     slices=args.slices,
                                                                     master=args.master,
                                                                     memory=args.memory,
-                                                                    dataset=os.path.splitext( os.path.basename( picklepath ) )[0],
-                                                                    verbose=(args.verbose and k_folds == 1),
+                                                                    dataset=os.path.splitext(os.path.basename(picklepath))[0],
+                                                                    verbose=args.verbose and args.psplit_training>0.0,
                                                                     log=run
                                                                   )
                             # Get best accuracy
@@ -451,7 +475,7 @@ def chopin2():
                         t1acc = time.time()
                         avg_accuracy = sum(accuracies)/len(accuracies)
                         avg_retraining = math.floor(sum(retrainings)/len(retrainings))
-                        message = 'Total elapsed time (accuracy) {}s\n'.format( int( t1acc - t0acc ) )
+                        message = 'Total elapsed time (testing) {}s\n'.format( int( t1acc - t0acc ) )
                         message += 'The maximum reached accuracy is {} after {} retrainings\n'.format( avg_accuracy, avg_retraining )
                         message += 'Stopped after {} retraining iterations\n'.format( len(accuracy)-1 )
                         fun.printlog( message, verbose=args.verbose, out=run )
@@ -467,7 +491,8 @@ def chopin2():
                                         "accuracy": avg_accuracy,
                                         "retraining": avg_retraining,
                                         "run": features_hash,
-                                        "features": comb_features
+                                        "features": comb_features,
+                                        "excluded": excluded_features
                                     }
                                 )
                             elif mapping[ group_size ][ 0 ][ "accuracy" ] == avg_accuracy:
@@ -476,7 +501,8 @@ def chopin2():
                                         "accuracy": avg_accuracy,
                                         "retraining": avg_retraining,
                                         "run": features_hash,
-                                        "features": comb_features
+                                        "features": comb_features,
+                                        "excluded": excluded_features
                                     }
                                 )
                         else:
@@ -485,7 +511,8 @@ def chopin2():
                                     "accuracy": avg_accuracy,
                                     "retraining": avg_retraining,
                                     "run": features_hash,
-                                    "features": comb_features
+                                    "features": comb_features,
+                                    "excluded": excluded_features
                                 }
                             ]
                         # Close log
@@ -493,7 +520,7 @@ def chopin2():
                             run.close()
                         # Cleanup
                         if args.cleanup:
-                            fun.cleanup( group_dir, datasetdir, args.dimensionality, args.levels, features_hash, k_folds=k_folds, spark=args.spark, skip_levels=True )
+                            fun.cleanup( group_dir, datasetdir, args.dimensionality, args.levels, features_hash, k_folds=args.crossv_k, spark=args.spark, skip_levels=True )
                         combinations_counter += 1
 
                     # Take track of the best result
@@ -510,7 +537,7 @@ def chopin2():
                         with open(summary_filepath, 'a+') as summary:
                             for run in mapping[group_size]:
                                 fun.printlog( 
-                                    '{}\t{}\t{}\t{}'.format(run["run"], group_size, run["retraining"], run["accuracy"]),
+                                    '{}\t{}\t{}\t{}\t{}'.format(run["run"], group_size, run["retraining"], run["accuracy"], ", ".join(run["excluded"])),
                                     out=summary
                                 )
         
@@ -529,7 +556,7 @@ def chopin2():
             fun.printlog( 'Selected features: {}'.format(fs_filepath), verbose=args.verbose )
 
     t1 = time.time()
-    fun.printlog('\nTotal elapsed time {}s'.format(int(t1 - t0)), verbose=args.verbose)
+    fun.printlog('Total elapsed time {}s'.format(int(t1 - t0)), verbose=args.verbose)
 
 if __name__ == '__main__':
     chopin2()
